=begin
(c) TIG 2010
    FixReversedFaceMaterials.rb
Usage:
    Plugins > Fix Reversed Face Materials
Select entities including any suspect faces [edges etc are ignored].
If any face has a back-material BUT it has no front-material it is assumed 
to have been made incorrectly and to be oriented the wrong way round, 
with the desired material applied to the back instead of the front.
So then the face is reversed and its front-material is made to match the 
back-material - including any UV-mapping appropriately corrected - 
the back-material is then removed.
If there is no selection you are asked if you want to process the whole 
model, if you reply "Yes" then absolutely everything is selected !
If the selection includes groups or components you are asked if you want to 
process these too - if you reply "Yes" then any faces inside them are also 
processed.  If there are 'nested' groups or components with these they are 
also processed through the nesting.
A closing dialog reports how many faces have been fixed.
To see the affect switch from Texture mode and view in MonoChrome mode, run 
the tool and you should see the changed faces flip - switch back to Texture 
mode and you should see no difference in appearance since all the materials 
should still be the same, but incorrectly oriented faces have been fixed.
Donations:
by PayPal to info@revitrev.org
Version:
1.0 20100806 First Issue.
1.1 20100811 Text deBableized.
=end
###
require 'sketchup.rb'
require 'deBabelizer.rb'
###
class FixReversedFaceMaterials

 ### TRANSLATOR ###
 def db(string)
  dir=File.join(File.dirname(__FILE__), "FixReversedFaceMaterials")
  locale=Sketchup.get_locale.upcase
  path=File.join(dir, ("FixReversedFaceMaterials"+locale+".lingvo"))
  if not File.exist?(path)
    return string
  else
    deBabelizer(string, path)
  end
 end#def

 def FixReversedFaceMaterials::db(string)
  dir=File.join(File.dirname(__FILE__), "FixReversedFaceMaterials")
  locale=Sketchup.get_locale.upcase
  path=File.join(dir, ("FixReversedFaceMaterials"+locale+".lingvo"))
  if not File.exist?(path)
    return string
  else
    deBabelizer(string, path)
  end
 end#if
 ############
 
  def activate
    @model=Sketchup.active_model
    sel=@model.selection
    sela=sel.to_a
    sel.clear
    if not sela[0] ### 6=YES 7=NO
      UI.beep
      yn=UI.messagebox( ( (db("There is NO selection!"))+"\n"+(db("Do you want to process ALL faces in the model?") ) ),MB_YESNO,"")
      self.deactivate(nil) if yn!=6
      return nil if yn!=6
      sela=@model.entities.to_a
    end#if
    @faces=[]; groups=[]; defs=[]; insts=0
    sela.each{|e|
      @faces<< e if e.class==Sketchup::Face
      groups<< e if e.class==Sketchup::Group
      defs<< e.definition if e.class==Sketchup::ComponentInstance and not defs.include?(e.definition)
      insts+=1 if e.class==Sketchup::ComponentInstance
    }
    if groups[1] ### 6=YES 7=NO
      yng=UI.messagebox( ( (db("There are"))+" "+groups.length.to_s+" "+(db("Groups in the selection!"))+"\n"+(db("Do you want to process any faces in them too?")) ),MB_YESNO,"")
    elsif groups[0]
      yng=UI.messagebox( ( (db("There is a Group in the selection!"))+"\n"+(db("Do you want to process any faces in it too?")) ),MB_YESNO,"")
    end#if
    ###
    if defs[1] ### 6=YES 7=NO
        ynd=UI.messagebox( ( (db("There are Instances of"))+" "+defs.length.to_s+" "+(db("Components in the selection!"))+"\n"+(db("Do you want to process any faces in them?")) ),MB_YESNO,"")
    elsif defs[0]
      if insts>1
        ynd=UI.messagebox( ( (db("There are Instances of a Component in the selection!"))+"\n"+(db("Do you want to process any faces in it?")) ),MB_YESNO,"")
      else
        ynd=UI.messagebox( ( (db("There is an Instance of a Component in the selection!"))+"\n"+(db("Do you want to process any faces in it?")) ),MB_YESNO,"")
      end#if
    end#if
    def do_group(g)
      g.entities.each{|e|
        @faces<< e if e.class==Sketchup::Face
        do_group(e) if e.class==Sketchup::Group
        do_def(e.definition) if e.class==Sketchup::ComponentInstance
      }
    end#def
    def do_def(d)
      d.entities.each{|e|
        @faces<< e if e.class==Sketchup::Face
        do_group(e) if e.class==Sketchup::Group
        do_def(e.definition) if e.class==Sketchup::ComponentInstance
      }
    end#def
    groups.each{|g|do_group(g)} if yng==6
    defs.each{|d| d.entities.each{|e|do_def(d) } } if ynd==6
    ### process faces
    count=self.process_faces(@faces)
    if count==1
      UI.messagebox(db("1 face fixed."))
    else
      UI.messagebox(count.to_s+" "+(db("faces fixed.")))
    end#if
    ###
    self.deactivate(nil)
    ###
  end#def
  
  def deactivate(view=nil)
    Sketchup.send_action("selectSelectionTool:")
    return nil
  end#if
  
  def process_faces(faces)
    @model.start_operation(db("Fix Reversed Face Materials"))
    texture_writer=Sketchup.create_texture_writer
    count=0
    faces.each{|face|
      if face.material==nil and face.back_material!=nil
        if face.back_material.texture==nil
          face.material=face.back_material
          face.back_material=nil ### remove this to leave back unchanged
          face.reverse!
        else
          samples = []
          samples << face.vertices[0].position			   ### 0,0 | Origin
		  samples << samples[0].offset(face.normal.axes.x) ### 1,0 | Offset Origin in X
	  	  samples << samples[0].offset(face.normal.axes.y) ### 0,1 | Offset Origin in Y
		  samples << samples[1].offset(face.normal.axes.y) ### 1,1 | Offset X in Y
		  xyz = [];uv = []### Arrays containing 3D and UV points.
		  uvh = face.get_UVHelper(true, true, texture_writer)
		  samples.each { |position|
			xyz << position ### XYZ 3D coordinates
			uvq = uvh.get_back_UVQ(position) ### UV 2D coordinates
			uv << self.flattenUVQ(uvq)
		  }
		  pts = [] ### Position texture.
		  (0..3).each { |i|
			 pts << xyz[i]
			 pts << uv[i]
		  }
          mat=face.back_material
          face.reverse!
		  face.position_material(mat, pts, true)
          face.back_material=nil ### remove this to leave back unchanged
        end#if
        count+=1
      end#if
    }
    @model.commit_operation
    return count
    ###
  end#def
  
  ### Get UV coordinates from UVQ matrix.
  def flattenUVQ(uvq)
	return Geom::Point3d.new(uvq.x / uvq.z, uvq.y / uvq.z, 1.0)
  end

end#class

def fixreversedfacematerials()
  Sketchup.active_model.select_tool(FixReversedFaceMaterials.new)
end#def

### add menu item etc #####################################################
if not file_loaded?(File.basename(__FILE__))
  ### make command
  text0 = (FixReversedFaceMaterials::db("Fix Reversed Face Materials"))
  text1 = (FixReversedFaceMaterials::db("Fix Reversed Face Materials"))
  text2 = (FixReversedFaceMaterials::db("Fix Reversed Face Materials"))+"..."
  cmd=UI::Command.new(text0){fixreversedfacematerials()}
  cmd.tooltip=text1
  cmd.status_bar_text=text2
  UI.menu("Plugins").add_item(cmd)
  ###
end#if
file_loaded(File.basename(__FILE__))
###
